package edu.ntnu.cikmr.arbitration;

import java.util.*;

import javax.swing.event.EventListenerList;

import edu.ntnu.cikmr.Framework;
import edu.ntnu.cikmr.arbitration.cbr.Conflict;
import edu.ntnu.cikmr.arbitration.workflow.*;
import edu.ntnu.cikmr.services.ServiceEvent;
import edu.ntnu.cikmr.services.ServiceListener;
import edu.ntnu.cikmr.services.peers.*;
import net.jxta.discovery.*;
import net.jxta.endpoint.*;
import net.jxta.peergroup.PeerGroup;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.util.JxtaBiDiPipe;

public class ConflictDiscoverer implements PeerListener, ServiceListener{
    private EventListenerList listenerList;
    private Hashtable<String, PeerWorkflowMonitor> peers;
    private PeersService peersService;
    private static final int INTERVAL = 10000;

    public ConflictDiscoverer(){
        listenerList = new EventListenerList();
        peers = new Hashtable<String,PeerWorkflowMonitor>();

        peersService = (PeersService)Framework.network.getServiceManager().get(PeersService.class);
        if(peersService == null) Framework.network.getServiceManager().add(peersService = new PeersService());
        peersService.addPeerListener(this);
        peersService.addServiceListener(this);
        if(peersService.isStarted()) loadPeers();
    }

    public void addConfictListener(ConflictListener l){
        listenerList.add(ConflictListener.class, l);
    }

    public void removeConfictListener(ConflictListener l){
        listenerList.remove(ConflictListener.class, l);
    }

    private void firePossibleConflict(ConflictEvent e){
        ConflictListener[] listeners = listenerList.getListeners(ConflictListener.class);
        for(int i = listeners.length - 1; i >= 0; i--) listeners[i].possibleConflict(e);
    }

    private synchronized void loadPeers(){
        Peer[] p = peersService.getPeers();
        for(int i = 0; i < p.length; i++) addPeer(p[i]);
    }

    private synchronized void clearPeers(){
        Enumeration<PeerWorkflowMonitor> enumeration = peers.elements();
        while(enumeration.hasMoreElements()){
            PeerWorkflowMonitor monitor = enumeration.nextElement();
            try{
                monitor.run = false;
                monitor.interrupt();
            } catch(Exception ex){}
        }
        peers.clear();
    }

    private synchronized void addPeer(Peer p){
        peers.put(p.getID(), new PeerWorkflowMonitor(p));
    }

    private synchronized void removePeer(Peer p){
        PeerWorkflowMonitor monitor = peers.remove(p.getID());
        try{
            monitor.run = false;
            monitor.interrupt();
        } catch(Exception ex){}
    }

    public void peerAdded(PeerEvent e){
        addPeer(e.getPeer());
    }

    public void peerRemoved(PeerEvent e){
        removePeer(e.getPeer());
    }

    public void serviceStarted(ServiceEvent e){
        loadPeers();
    }

    public void serviceStopped(ServiceEvent e){
        clearPeers();
    }

    private class PeerWorkflowMonitor extends Thread{
        private Peer peer;
        private boolean run;

        private PeerWorkflowMonitor(Peer p){
            peer = p;
            run = true;
            start();
        }

        public void run(){
            while(run){
                try{
                    PeerWorkflowContainer container = getPeerWorkflows(peer);
                    checkForConflicts(container);
                } catch(Exception ex){
                    ex.printStackTrace();
                }
                try{
                    Thread.sleep(INTERVAL);
                } catch(Exception ex){}
            }
        }

    }

    private void checkForConflicts(PeerWorkflowContainer remoteContainer){
        if(remoteContainer == null || !remoteContainer.activityIterator().hasNext()) return;

        PeerWorkflowContainer localContainer = Framework.scheduler.getWorkflowContainer();

        List<CaseContainer> cases = new ArrayList<CaseContainer>();

        traverseWidth(cases, localContainer, remoteContainer);
        traverseAll(cases, localContainer.activityIterator(), remoteContainer);

        ConflictEvent conflicts = getConflicts(cases);
        if(conflicts != null){
            // Possible conflicts have been found. What to do?
            conflicts.setPeerID(remoteContainer.getId());
            conflicts.setPeerName(remoteContainer.getName());
            firePossibleConflict(conflicts);
        }
    }

    private ConflictEvent getConflicts(List<CaseContainer> cases){
        ConflictEvent event = null;
        Conflict[] conflicts = null;
        CaseContainer container = null;
        Iterator<CaseContainer> iterator = cases.iterator();

        while(iterator.hasNext() && conflicts == null){
            container = iterator.next();
            conflicts = Framework.arbitration.getCBR().checkForConflict(container);
        }
        if(conflicts != null){
            event = new ConflictEvent(this);
            event.setContainer(container);
            event.setConflicts(conflicts);
            return event;
        }
        return null;
    }

    private void traverseWidth(List<CaseContainer> cases, PeerWorkflowContainer local, PeerWorkflowContainer remote){
        Iterator<Activity> localIterator = local.activityIterator();

        while(localIterator.hasNext()){
            Activity localA = localIterator.next();
            Iterator<Activity> remoteIterator = remote.activityIterator();
            while(remoteIterator.hasNext()){
                Activity remoteA = remoteIterator.next();
                if(localA instanceof Goal && remoteA instanceof Goal) traverseGoalWidth(cases, (Goal)localA, (Goal)remoteA);
            }
        }
    }

    private void traverseGoalWidth(List<CaseContainer> cases, Goal local, Goal remote){
        Iterator<Activity> localIterator = local.activityIterator();
        Iterator<Activity> remoteIterator = remote.activityIterator();

        CaseContainer container = new CaseContainer(local, remote);
        if(!cases.contains(container)){
            cases.add(container);
        }

        while(localIterator.hasNext() && remoteIterator.hasNext()){
            container = new CaseContainer(localIterator.next(), remoteIterator.next());
            if(!cases.contains(container)){
                cases.add(container);
            }
            if(container.getLocal() instanceof Goal && container.getRemote() instanceof Goal){
                traverseGoalWidth(cases, (Goal)container.getLocal(), (Goal)container.getRemote());
            }
        }
    }

    private void traverseAll(List<CaseContainer> cases, Iterator<Activity> local, PeerWorkflowContainer remote){

        while(local.hasNext()){
            Activity activity = local.next();
            traverseAll(cases, activity,  remote.activityIterator());
            if(activity instanceof Goal) traverseAll(cases, ((Goal)activity).activityIterator(), remote);
        }
    }

    private void traverseAll(List<CaseContainer> cases, Activity local, Iterator<Activity> remote){
        while(remote.hasNext()){
            Activity item = remote.next();
            CaseContainer container = new CaseContainer(local, item);
            if(!cases.contains(container)){
                cases.add(container);
            }
            if(item instanceof Goal) traverseAll(cases, local, ((Goal)item).activityIterator());
        }
    }

    private PeerWorkflowContainer getPeerWorkflows(Peer peer){
        JxtaBiDiPipe pipe = null;
        try {
            PeerGroup peerGroup = Framework.network.getPeerGroup();

            // Create and connect to resource pipe
            //PipeID pipeID = IDFactory.newPipeID(peerGroup.getPeerGroupID(), new String("Scheduler:" + peer.getID()).getBytes());
            //PipeAdvertisement pipeAdvertisement = PipeUtilities.createPipeAdvertisement(pipeID.toString(), PipeService.UnicastType);
            //PipeAdvertisement pipeAdvertisement = PipeUtilities.createPipeAdvertisement("Scheduler:" + peer.getID(), PipeService.UnicastType);

            PipeAdvertisement pipeAdvertisement = getSchedulerPipeAdvertisement(peerGroup, peer);

            pipe = new JxtaBiDiPipe();
            pipe.setReliable(true);
            pipe.connect(Framework.network.getPeerGroup(), pipeAdvertisement, 5000);
            // Send the invoke method request (Should probably have used RMI)
            pipe.sendMessage(createWorkflowsRequestMessage());
            // Recieve response
            Message responseMessage = pipe.getMessage(10000);
            // Close pipe
            pipe.close();
            // Extract response from message
            return getPeerWorkflows(responseMessage);
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        try{
            pipe.close();
        }catch(Exception ex){}
        return null;
    }

    private PipeAdvertisement getSchedulerPipeAdvertisement(PeerGroup peerGroup, Peer peer) throws Exception{
        PipeAdvertisementFinder finder = new PipeAdvertisementFinder(peerGroup, peer);
        return finder.get();
    }

    private class PipeAdvertisementFinder extends Thread implements DiscoveryListener{
        PeerGroup peerGroup;
        Peer peer;
        PipeAdvertisement pipeAdvertisement;
        int wait = 10000;
        Object lock;

        private PipeAdvertisementFinder(PeerGroup peerGroup, Peer peer){
            this.peerGroup = peerGroup;
            this.peer = peer;
            lock = new Object();
            pipeAdvertisement = null;
            start();
        }

        public void run(){
            synchronized(lock){
                peerGroup.getDiscoveryService().getRemoteAdvertisements("urn:jxta:" + peer.getID(), DiscoveryService.ADV, "Name", "Scheduler:" + peer.getID(), 10, this);
                try{
                    Thread.sleep(wait);
                } catch(Exception ex){}
                lock.notify();
            }
        }

        public void discoveryEvent(DiscoveryEvent discoveryEvent){
            Enumeration enumeration = discoveryEvent.getSearchResults();
            while(enumeration.hasMoreElements()){
                try{
                    PipeAdvertisement adv = (PipeAdvertisement)enumeration.nextElement();
                    if(adv.getName().equalsIgnoreCase("Scheduler:" + peer.getID())){
                        pipeAdvertisement = adv;
                        try{
                            this.interrupt();
                        } catch(Exception ex){}
                        return;
                    }
                } catch(Exception ex){}
            }
        }

        public PipeAdvertisement get() throws Exception{
            synchronized(lock){
                try{
                    lock.wait();
                } catch(Exception ex){}
                if(pipeAdvertisement == null) throw new Exception("Unable to find scheduler pipe for given peer. (" + peer.getName() + ")");
                return pipeAdvertisement;
            }
        }

    }

    private PeerWorkflowContainer getPeerWorkflows(Message message){
        ByteArrayMessageElement byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("peerID");
        String peerID = new String(byteArrayElement.getBytes());
        byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("command");
        String command = new String(byteArrayElement.getBytes());
        try{
            if(command.equalsIgnoreCase("Scheduler:Workflows")){
                byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("xml");
                String xml = new String(byteArrayElement.getBytes());
                PeerWorkflowContainer container = new PeerWorkflowContainer();
                try{
                    container.fromXML(xml);
                    return container;
                } catch(Exception ex){
                    ex.printStackTrace();
                }

            }
        } catch(Exception ex){
            ex.printStackTrace();
       }
       return null;
    }

    private Message createWorkflowsRequestMessage(){
        PeerGroup peerGroup = Framework.network.getPeerGroup();

        StringMessageElement nameMessageElement = new StringMessageElement("peerName", peerGroup.getPeerName(), null);
        StringMessageElement idMessageElement = new StringMessageElement("peerID", peerGroup.getPeerID().getUniqueValue().toString(), null);
        StringMessageElement commandMessageElement = new StringMessageElement("command", "Scheduler:Workflows", null);

        Message message = new Message();
        message.addMessageElement(nameMessageElement);
        message.addMessageElement(idMessageElement);
        message.addMessageElement(commandMessageElement);

        return message;
    }

}
