package edu.ntnu.cikmr.scheduling;

import java.io.IOException;
import java.net.SocketTimeoutException;
import java.util.*;

import javax.swing.event.EventListenerList;

import edu.ntnu.cikmr.Framework;
import edu.ntnu.cikmr.arbitration.workflow.PeerWorkflowContainer;
import edu.ntnu.cikmr.services.AbstractService;
import edu.ntnu.cikmr.services.ServiceManager;
import edu.ntnu.cikmr.workflowenactment.Workflow;
import edu.ntnu.cikmr.workflowenactment.WorkflowEvent;
import net.jxta.discovery.DiscoveryService;
import net.jxta.document.Advertisement;
import net.jxta.endpoint.*;
import net.jxta.peergroup.PeerGroup;
import net.jxta.pipe.PipeService;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.util.*;

public class Scheduler extends AbstractService{
    private List<Workflow> workflows;
    private EventListenerList listenerList;
    private boolean started;
    private JxtaServerPipe serverPipe;
    private ServerPipeRunner serverPipeRunner;
    private PeerGroup peerGroup;

    public Scheduler(){
        workflows = Collections.synchronizedList(new ArrayList<Workflow>());
        listenerList = new EventListenerList();
    }

    public Workflow addWorkflow(String xmlDefinition){
        if(!started) return null;
        try{
            Workflow workflow = new Workflow(xmlDefinition);
            workflows.add(workflow);
            fireWorkflowAdded(workflow);
            return workflow;
        } catch(Exception ex){
            return null;
        }
    }

    public boolean removeWorkflow(Workflow workflow){
        synchronized(workflows){
            ListIterator<Workflow> iterator = workflows.listIterator();
            while (iterator.hasNext()) {
                if (iterator.next() == workflow) {
                    iterator.remove();
                    workflow.release();
                    fireWorkflowRemoved(workflow);
                    return true;
                }
            }
            return false;
        }
    }

    private void removeAllWorkflows(){
        synchronized(workflows){
            ListIterator<Workflow> iterator = workflows.listIterator();
            while(iterator.hasNext()){
                Workflow workflow = iterator.next();
                iterator.remove();
                workflow.release();
            }
        }
    }

    public Workflow getWorkflow(int index){
        return workflows.get(index);
    }

    public Workflow getWorkflow(String id){
        synchronized(workflows){
            Iterator<Workflow> iterator = workflows.iterator();
            while(iterator.hasNext()){
                Workflow workflow = iterator.next();
                if(workflow.getID().equals(id)) return workflow;
            }
        }
        return null;
    }

    public int getWorkflowCount(){
        return workflows.size();
    }

    public void addSchedulerListener(SchedulerListener l){
        listenerList.add(SchedulerListener.class, l);
    }

    public void removeSchedulerListener(SchedulerListener l){
        listenerList.remove(SchedulerListener.class, l);
    }

    private void fireWorkflowAdded(Workflow workflow){
        WorkflowEvent event = new WorkflowEvent(this, workflow);
        SchedulerListener[] listeners = listenerList.getListeners(SchedulerListener.class);
        for(int i = listeners.length - 1; i >= 0; i--) listeners[i].workflowAdded(event);
    }

    private void fireWorkflowRemoved(Workflow workflow){
        WorkflowEvent event = new WorkflowEvent(this, workflow);
        SchedulerListener[] listeners = listenerList.getListeners(SchedulerListener.class);
        for(int i = listeners.length - 1; i >= 0; i--) listeners[i].workflowRemoved(event);
    }

    public void init(ServiceManager manager) {
    }

    public void start(PeerGroup peerGroup) throws IOException {
        if(serverPipeRunner != null && serverPipeRunner.run) return;
        started = false;
        this.peerGroup = peerGroup;
        try{
            Enumeration advs = peerGroup.getDiscoveryService().getLocalAdvertisements(DiscoveryService.ADV, "Name", "Scheduler:*");
            while(advs.hasMoreElements()){
                Advertisement adv = (Advertisement)advs.nextElement();
                peerGroup.getDiscoveryService().flushAdvertisement(adv);
            }
        } catch(Exception ex){
            ex.printStackTrace();
        }

        try{
            //PipeID pipeID = IDFactory.newPipeID(peerGroup.getPeerGroupID(), peerGroup.getPeerName().getBytes());
            //PipeAdvertisement pipeAdvertisement = PipeUtilities.createNewPipeAdvertisement(peerGroup, PipeService.UnicastType);
            PipeAdvertisement pipeAdvertisement = PipeUtilities.createNewPipeAdvertisement(peerGroup, PipeService.UnicastType);
            pipeAdvertisement.setName("Scheduler:" + peerGroup.getPeerID().getUniqueValue().toString());

            peerGroup.getDiscoveryService().publish(pipeAdvertisement);
            serverPipe = new JxtaServerPipe(peerGroup, pipeAdvertisement, 20, 2000);

            serverPipeRunner = new ServerPipeRunner();
        }catch(Exception ex){
            ex.printStackTrace();
            throw new IOException("Unable to start server pipe.");
        }
        started = true;
    }

    public void stop() throws IOException {
        if(serverPipeRunner != null && serverPipeRunner.run){
            started = false;
            serverPipeRunner.run = false;

            removeAllWorkflows();

            PipeAdvertisement adv = serverPipe.getPipeAdv();
            try{
                serverPipeRunner.interrupt();
                serverPipeRunner.join();
            } catch(Exception ex){}
            try{
                serverPipe.close();
            } catch(Exception ex){
                ex.printStackTrace();
            }
            serverPipe = null;
            serverPipeRunner = null;
            peerGroup.getDiscoveryService().flushAdvertisement(adv);
        }
    }

    public boolean isStarted() {
        return started;
    }

    private class ServerPipeRunner extends Thread{
        private boolean run;
        private JxtaBiDiPipe pipe;

        public ServerPipeRunner(){
            run = true;
            start();
        }

        public void run(){
            while(run){
                try {
                    pipe = serverPipe.accept();
                } catch (SocketTimeoutException ex) {
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
                if(pipe != null){
                    new ConnectionHandler(pipe);
                    pipe = null;
                }
            }
        }

    }

    private class ConnectionHandler extends Thread{
        private JxtaBiDiPipe pipe;

        private ConnectionHandler(JxtaBiDiPipe pipe){
            this.pipe = pipe;
            start();
        }

        public void run(){
            try {
                Message requestMessage = pipe.getMessage(10000);
                Message responseMessage = createResponse(requestMessage);
                if(responseMessage != null) pipe.sendMessage(responseMessage);
                try {
                    pipe.close();
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }

    }

    private Message createResponse(Message message){
        ByteArrayMessageElement byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("peerID");
        String peerID = new String(byteArrayElement.getBytes());
        byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("command");
        String command = new String(byteArrayElement.getBytes());
        try{
            if(command.equalsIgnoreCase("Scheduler:Workflows")){
                return createWorkflowsMessage();
            }
            return null;
        } catch(Exception ex){
            ex.printStackTrace();
            return null;
        }
    }

    private Message createWorkflowsMessage(){
        PeerGroup peerGroup = Framework.network.getPeerGroup();

        StringMessageElement nameMessageElement = new StringMessageElement("peerName", peerGroup.getPeerName(), null);
        StringMessageElement idMessageElement = new StringMessageElement("peerID", peerGroup.getPeerID().getUniqueValue().toString(), null);
        StringMessageElement commandMessageElement = new StringMessageElement("command", "Scheduler:Workflows", null);
        StringMessageElement xmlMessageElement = new StringMessageElement("xml", createWorkflowsXML(), null);

        Message message = new Message();
        message.addMessageElement(nameMessageElement);
        message.addMessageElement(idMessageElement);
        message.addMessageElement(commandMessageElement);
        message.addMessageElement(xmlMessageElement);

        return message;
    }

    private String createWorkflowsXML(){
        if(!started) return null;
        String xml = "<workflows peerName=\"" + peerGroup.getPeerName() + "\" peerID=\"" + peerGroup.getPeerID().getUniqueValue().toString() + "\">\n";
        Iterator<Workflow> iterator = workflows.iterator();
        while(iterator.hasNext()){
            Workflow workflow = iterator.next();
            if(workflow.isExecutedLocally() && workflow.isExecuting()) xml += workflow.toTypeXML() + "\n";
        }
        xml += "</workflows>";
        return xml;
    }

    public PeerWorkflowContainer getWorkflowContainer(){
        PeerWorkflowContainer container = new PeerWorkflowContainer();
        try{
            String xml = createWorkflowsXML();
            container.fromXML(xml);
            return container;
        } catch(Exception ex){
            ex.printStackTrace();
            return null;
        }
    }


}
