package edu.ntnu.cikmr.services.context;

import java.io.*;
import java.rmi.*;
import java.util.*;

import dk.pervasive.jcaf.*;
import dk.pervasive.jcaf.impl.*;
import dk.pervasive.jcaf.util.*;
import edu.ntnu.cikmr.*;
import edu.ntnu.cikmr.log.*;
import edu.ntnu.cikmr.services.*;
import edu.ntnu.cikmr.services.context.entities.*;
import edu.ntnu.cikmr.services.peers.*;
import net.jxta.document.*;
import net.jxta.peergroup.*;
import edu.ntnu.cikmr.services.context.entities.Person;

public class ContextService extends AbstractService implements ServiceListener, PeerListener, EntityListener{
    private ServiceManager serviceManager;
    private ContextServiceImpl contextService;
    private boolean isStarted;
    private PeersService peersService;
    private Map peerContextServices;
    private RemoteEntityListenerImpl remoteEntityListener;

    public ContextService(){
        peerContextServices = Collections.synchronizedMap(new Hashtable());
        try{
            remoteEntityListener = new RemoteEntityListenerImpl();
            remoteEntityListener.addEntityListener(this);
        } catch(RemoteException ex){
            ex.printStackTrace();
        }
  }

//- Context operations ---------------------------------------------------------

    public void addContextItem(String id, Relationship r, ContextItem item){
        try{
            contextService.addContextItem(id, r, item);
        } catch(RemoteException ex){
            ex.printStackTrace();
        }
        System.out.println("ContextItem added: " + item.toXML());
    }

    public void addEntity(Entity e){
        try{
            contextService.addEntity(e);
        } catch(RemoteException ex){
            ex.printStackTrace();
        }
        System.out.println("Entity added: " + e.toXML());
    }

    public Entity[] getAllEntitiesByType(Class c){
        return getAllEntitiesByType(c, true, true);
    }

    public Entity[] getAllLocalEntitiesByType(Class c){
        return getAllEntitiesByType(c, true, false);
    }

    public Entity[] getAllRemoteEntitiesByType(Class c){
        return getAllEntitiesByType(c, false, true);
    }

    private Entity[] getAllEntitiesByType(Class c, boolean includeLocal, boolean includeRemote){
        Entity[] e0 = null;

        if(includeLocal){
            try{
                Entity[] e1 = contextService.getAllEntitiesByType(c);
                e0 = joinArrays(e0, e1);
            }catch(Exception ex){}
        }

        if(includeRemote){
            synchronized(peerContextServices){
                Iterator i = peerContextServices.values().iterator();
                while(i.hasNext()){
                    try{
                        Entity[] e1 = ((ContextServiceMonitor)i.next()).getContextService().getAllEntitiesByType(c);
                        e0 = joinArrays(e0, e1);
                    } catch(Exception ex){}
                }
            }
        }
        return e0;
    }

//------------------------------------------------------------------------------

    private Entity[] joinArrays(Entity[] a1, Entity[] a2){
        if(a1 == null) return a2;
        if(a2 == null) return a1;
        Entity[] a3 = new Entity[a1.length + a2.length];
        System.arraycopy(a1, 0, a3, 0, a1.length);
        System.arraycopy(a2, 0, a3, a1.length, a2.length);
        return a3;
    }

    public void init(ServiceManager manager) {
        serviceManager = manager;
        serviceManager.addServiceListener(this);
    }

    private void startContextConnectionsToPeers(){
        Peer[] peers = peersService.getPeers();
        for (int i = 0; i < peers.length; i++) new PeerContextServiceAdder(peers[i]);
    }

    private void endContextConnectionsToPeers(){
        synchronized(peerContextServices){
            Iterator i = peerContextServices.values().iterator();
            while(i.hasNext()){
                ContextServiceMonitor monitor = (ContextServiceMonitor)i.next();
                new PeerContextServiceRemover(monitor.getPeer());
            }
        }
    }

    public void addListener(){
        try{
            contextService.addEntityListener(remoteEntityListener, Person.class);
        } catch(RemoteException ex){
        }
    }

    public synchronized void start(PeerGroup peerGroup) throws IOException {
        isStarted = true;
        String uniqueValue = peerGroup.getPeerID().getUniqueValue().toString();
        contextService = new ContextServiceImpl(uniqueValue);

        fireServiceStarted(this);
        Framework.log.logMessage(this, Log.LEVEL_DEBUG, "Started (" + contextService.getURI() + ").");

        if(peersService == null) peersService = (PeersService)serviceManager.get(PeersService.class);
        if(peersService != null && peersService.isStarted()){
            peersService.addPeerListener(this);
            startContextConnectionsToPeers();
        }
    }

    public synchronized void stop() throws IOException {
        isStarted = false;
        contextService.removeEntityListener(remoteEntityListener, "");
        contextService = null;
        fireServiceStopped(this);
        Framework.log.logMessage(this, Log.LEVEL_DEBUG, "Stopped.");

        if(peersService != null){
            peersService.removePeerListener(this);
            peersService = null;
            endContextConnectionsToPeers();
        }
    }

    public boolean isStarted() {
        return isStarted;
    }

    public void serviceStarted(ServiceEvent e){
        if(peersService == e.getService() && isStarted()){
            peersService.addPeerListener(this);
            startContextConnectionsToPeers();
        }else if(peersService == null && e.getService() instanceof PeersService){
            peersService = (PeersService)e.getService();
            if(isStarted()){
                peersService.addPeerListener(this);
                startContextConnectionsToPeers();
            }
        }
    }

    public void serviceStopped(ServiceEvent e){
        if(peersService == e.getService()){
            peersService.removePeerListener(this);
            peersService = null;
            endContextConnectionsToPeers();
        }
    }

    private String getIPAddress(Peer peer){
        XMLDocument doc = (XMLDocument)peer.getAdvertisement().getDocument(MimeMediaType.XMLUTF8);
        return findIPAddress(doc);
    }

    private String findIPAddress(Element e){
        if(e.getKey().toString().equalsIgnoreCase("EA")){
            String value = e.getValue().toString();
            if(value.toLowerCase().startsWith("tcp://")){
                value = value.substring("tcp://".length());
                int index = value.indexOf(':');
                return index == -1 ? value : value.substring(0, index);
            }
        }
        Enumeration enumeration = e.getChildren();
        while(enumeration.hasMoreElements()){
            String ret = findIPAddress((Element)enumeration.nextElement());
            if(ret != null) return ret;
        }
        return null;
    }

    public void peerAdded(PeerEvent e){
        new PeerContextServiceAdder(e.getPeer());
    }

    public void peerRemoved(PeerEvent e){
        new PeerContextServiceRemover(e.getPeer());
    }

    public void contextChanged(ContextEvent contextEvent){
        System.out.println("-- contextChanged event --");
        System.out.println(contextEvent.getSource().getClass().getName());
        System.out.println(contextEvent.getEntity().getClass().getName());
        System.out.println(contextEvent.getEventType());
        System.out.println("---------------------------");
    }

    private class PeerContextServiceAdder extends Thread{
        private Peer peer;

        public PeerContextServiceAdder(Peer peer){
            this.peer = peer;
            this.start();
        }

        public void run(){
            if(contextService != null){
                String ip = getIPAddress(peer);
                if(ip != null){
                    try {
                        String context = "rmi://" + ip + "/" + peer.getID();
                        ContextServiceMonitor monitor = new ContextServiceMonitor(context, peer);

                        if(contextService.addContextService(monitor.getContextService())){
                            peerContextServices.put(peer.getID(), monitor);
/*
                            try{
                                RemoteEntityListenerImpl rel = new RemoteEntityListenerImpl();
                                rel.addEntityListener(monitor);
                                monitor.getContextService().addEntityListener(rel, Person.class);
                            } catch(Exception ex){}
*/
                            Framework.log.logMessage(ContextService.this, Log.LEVEL_DEBUG, "Added peer context service (" + peer.getName() + ").");
                        }else{
                            Framework.log.logMessage(ContextService.this, Log.LEVEL_DEBUG, "Unable to add peer context service (" + peer.getName() + ").");
                        }
                    } catch(Exception ex) {
                        ex.printStackTrace();
                    }
                }
            }
        }
    }

    private class PeerContextServiceRemover extends Thread{
        private Peer peer;

        public PeerContextServiceRemover(Peer peer){
            this.peer = peer;
            this.start();
        }

        public void run(){
            if(contextService != null){
                try {
                    ContextServiceMonitor monitor = (ContextServiceMonitor)peerContextServices.remove(peer.getID());
                    if(monitor != null){

                        if(contextService.removeContextService(monitor.getContextService())){
                            Framework.log.logMessage(this, Log.LEVEL_DEBUG, "Removed peer context service (" + peer.getName() + ").");
                        }else{
                            Framework.log.logMessage(this, Log.LEVEL_DEBUG, "Unable to remove peer context service (" + peer.getName() + ").");
                        }
                    }
                } catch (RemoteException ex) {
                    ex.printStackTrace();
                }
            }
        }
    }

    private class ContextServiceMonitor extends AbstractContextClient implements EntityListener, Serializable{
        private transient Peer peer;

        public ContextServiceMonitor(String uri, Peer peer) throws Exception{
            super(uri);
            this.peer = peer;
        }

        public Peer getPeer(){
            return peer;
        }

        public dk.pervasive.jcaf.ContextService getContextService(){
            return super.getContextService();
        }

        public void monitor(String string) throws RemoteException {
            System.out.println(peer.getName() + " " + "monitor: " + string);
        }

        public void run() {
        }

        public void contextChanged(ContextEvent contextEvent){
            System.out.println("-- contextChanged event --");
            System.out.println(contextEvent.getSource().getClass().getName());
            System.out.println(contextEvent.getEntity().getClass().getName());
            System.out.println(contextEvent.getEventType());
            System.out.println("---------------------------");
        }

    }

}
