package edu.ntnu.cikmr.services.resources;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import javax.xml.datatype.*;
import javax.xml.parsers.*;

import edu.ntnu.cikmr.services.*;
import net.jxta.endpoint.*;
import net.jxta.peergroup.*;
import net.jxta.pipe.*;
import net.jxta.protocol.*;
import net.jxta.util.*;
import org.w3c.dom.*;
import java.net.SocketTimeoutException;
import edu.ntnu.cikmr.log.Log;
import edu.ntnu.cikmr.Framework;

public abstract class AbstractResourceService extends AbstractService {
    private PeerGroup resourcePeerGroup;
    private JxtaServerPipe resourceServerPipe;
    private ResourceServerPipeRunner resourceServerPipeRunner;
    private PipeAdvertisement resourcePipeAdvertisement;
    private String uniqueID;
    private String resourceType;

    public abstract ResourceDescription getResourceDescription();

    public final void startResource(PeerGroup peerGroup) throws InstantiationException{
        try{
            resourcePeerGroup = peerGroup;
            resourceType = getResourceDescription().getType().getLocalName();
            uniqueID = ResourceUtils.generateUniqueID();
            resourcePipeAdvertisement = PipeUtilities.createNewPipeAdvertisement(resourcePeerGroup, PipeService.UnicastType);
            resourcePipeAdvertisement.setName("Resource:" + peerGroup.getPeerID().getUniqueValue().toString() + ":" + resourceType + ":" + uniqueID);
            resourcePeerGroup.getDiscoveryService().publish(resourcePipeAdvertisement);
            resourceServerPipe = new JxtaServerPipe(resourcePeerGroup, resourcePipeAdvertisement, 20, 2000);
            resourceServerPipeRunner = new ResourceServerPipeRunner();
            Framework.log.logMessage(this, Log.LEVEL_DEBUG, "Started. (" + uniqueID + ")");
        }catch(Exception ex){
            ex.printStackTrace();
            throw new InstantiationException("Unable to set up resource communication pipe.");
        }
    }

    public final void stopResource(){
        if(resourceServerPipeRunner == null) return;
        resourceServerPipeRunner.run = false;
        try {
            resourceServerPipeRunner.join(5000);
        } catch (Exception ex) {}
        resourceServerPipeRunner = null;
        try {
            resourceServerPipe.close();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        resourceServerPipe = null;
        try{
            resourcePeerGroup.getDiscoveryService().flushAdvertisement(resourcePipeAdvertisement);
        } catch(Exception ex){
            ex.printStackTrace();
        }
        resourcePeerGroup = null;
        Framework.log.logMessage(this, Log.LEVEL_DEBUG, "Stopped. (" + uniqueID + ")");
    }

    private class ResourceServerPipeRunner extends Thread{
        private boolean run;
        private JxtaBiDiPipe pipe;

        public ResourceServerPipeRunner(){
            run = true;
            start();
        }

        public void run(){
            Framework.log.logMessage(this, Log.LEVEL_DEBUG, "Started server pipe. (" + uniqueID + ")");
            while(run){
                try {
                    pipe = resourceServerPipe.accept();
                } catch (SocketTimeoutException ex) {
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
                if(pipe != null){
                    new ResourceConnectionHandler(pipe);
                    pipe = null;
                }
            }
            Framework.log.logMessage(this, Log.LEVEL_DEBUG, "Stopped server pipe. (" + uniqueID + ")");
        }

    }

    private class ResourceConnectionHandler extends Thread{
        private JxtaBiDiPipe pipe;

        private ResourceConnectionHandler(JxtaBiDiPipe pipe){
            this.pipe = pipe;
            start();
        }

        public void run(){
            try {
                Framework.log.logMessage(AbstractResourceService.this, Log.LEVEL_DEBUG, "Recieved message. (" + uniqueID + ")");
                Message requestMessage = pipe.getMessage(10000);
                Message responseMessage = respond(requestMessage);
                Framework.log.logMessage(AbstractResourceService.this, Log.LEVEL_DEBUG, "Responding. (" + uniqueID + ")");
                pipe.sendMessage(responseMessage);
                try {
                    pipe.close();
                } catch (Exception ex) {}
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }

    }

    private Message respond(Message requestMessage) throws InvocationTargetException{
        ByteArrayMessageElement byteArrayElement = (ByteArrayMessageElement)requestMessage.getMessageElement("command");
        String command = new String(byteArrayElement.getBytes());

        if(command.equalsIgnoreCase("Resource:InvokeMethods")){
            byteArrayElement = (ByteArrayMessageElement)requestMessage.getMessageElement("xml");
            String xml = new String(byteArrayElement.getBytes());

            String resultXML = invokeMethods(xml);
            return createResultMessage(resultXML);
        }else if(command.equalsIgnoreCase("Resource:GetDescription")){
            String descriptionXML = ResourceDescription.toXML(getResourceDescription());
            return createDescriptionMessage(descriptionXML);
        }
        throw new InvocationTargetException(new NullPointerException("Could not find invoke methods message elements."), "Invalid formatted message.");
    }

    private Message createDescriptionMessage(String xml) throws InvocationTargetException{
        StringMessageElement nameMessageElement = new StringMessageElement("peerName", resourcePeerGroup.getPeerName(), null);
        StringMessageElement idMessageElement = new StringMessageElement("peerID", resourcePeerGroup.getPeerID().getUniqueValue().toString(), null);
        StringMessageElement pipeIDMessageElement = new StringMessageElement("pipeID", "", null);
        StringMessageElement commandMessageElement = new StringMessageElement("command", "Resource:Description", null);
        StringMessageElement xmlMessageElement = new StringMessageElement("xml", xml, null);

        Message message = new Message();
        message.addMessageElement(nameMessageElement);
        message.addMessageElement(idMessageElement);
        message.addMessageElement(pipeIDMessageElement);
        message.addMessageElement(commandMessageElement);
        message.addMessageElement(xmlMessageElement);

        return message;
    }

    private Message createResultMessage(String xml) throws InvocationTargetException{
        StringMessageElement nameMessageElement = new StringMessageElement("peerName", resourcePeerGroup.getPeerName(), null);
        StringMessageElement idMessageElement = new StringMessageElement("peerID", resourcePeerGroup.getPeerID().getUniqueValue().toString(), null);
        StringMessageElement pipeIDMessageElement = new StringMessageElement("pipeID", "", null);
        StringMessageElement commandMessageElement = new StringMessageElement("command", "Resource:InvokedMethodsResult", null);
        StringMessageElement xmlMessageElement = new StringMessageElement("xml", xml, null);

        Message message = new Message();
        message.addMessageElement(nameMessageElement);
        message.addMessageElement(idMessageElement);
        message.addMessageElement(pipeIDMessageElement);
        message.addMessageElement(commandMessageElement);
        message.addMessageElement(xmlMessageElement);

        return message;
    }

    public String invokeMethods(String xml) throws InvocationTargetException{
        MethodInvocation[] methodInvocations = getMethodInvocations(xml);
        if(methodInvocations != null && methodInvocations.length > 0){
            for (int i = 0; i < methodInvocations.length; i++) {
                invokeMethod(methodInvocations[i]);
            }
        }
        return createXMLResult(methodInvocations);
    }

    private MethodInvocation[] getMethodInvocations(String xml) throws InvocationTargetException{
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(new ByteArrayInputStream(xml.getBytes()));

            NodeList nodeList = doc.getChildNodes();
            for (int i = 0; i < nodeList.getLength(); i++) {
                Node node = nodeList.item(i);
                if(node.getNodeName().equalsIgnoreCase("invokeMethods")){
                    return getMethodInvocations(node);
                }
            }
        }catch(Exception ex){
            throw new InvocationTargetException(ex, "Unable to invoke methods.");
        }
        throw new InvocationTargetException(new NullPointerException("Unable to find invoke methods node."), "Unable to invoke methods.");
    }

    private MethodInvocation[] getMethodInvocations(Node invokeMethodsNode) throws InvocationTargetException{
        ArrayList<MethodInvocation> methodInvocations = new ArrayList<MethodInvocation>();

        NodeList nodeList = invokeMethodsNode.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node node = nodeList.item(i);
            if(node.getNodeName().equalsIgnoreCase("method")){
                methodInvocations.add(getMethodInvocation(node));
            }
        }
        if(methodInvocations.size() == 0) throw new InvocationTargetException(new NullPointerException("No methods found"), "Unable to invoke methods.");
        return methodInvocations.toArray(new MethodInvocation[methodInvocations.size()]);
    }

    private MethodInvocation getMethodInvocation(Node methodNode) throws InvocationTargetException{
        try {

            Node nameNode = methodNode.getAttributes().getNamedItem("name");
            ResourceMethod method = getResourceDescription().getMethod(nameNode.getNodeValue());
            Object[] parameters = getMethodParameters(method, methodNode);

            MethodInvocation methodInvocation = new MethodInvocation();
            methodInvocation.method = method;
            methodInvocation.parameters = parameters;

            return methodInvocation;
        } catch (Exception ex) {
            throw new InvocationTargetException(ex, "Unable to extract method.");
        }
    }

    private Object[] getMethodParameters(ResourceMethod method, Node methodNode) throws DatatypeConfigurationException{
        String[] parameterTypes = method.getParameters();
        int parameterTypeLength = parameterTypes == null ? 0 : parameterTypes.length;

        if(parameterTypeLength == 0) return null;
        ArrayList<String> parametersArrayList = new ArrayList<String>(parameterTypeLength);

        NodeList nodeList = methodNode.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node node = nodeList.item(i);
            if(node.getNodeName().equalsIgnoreCase("parameter")){
                NamedNodeMap attributes = node.getAttributes();
                Node valueNode = attributes.getNamedItem("value");
                parametersArrayList.add(valueNode.getNodeValue());
            }
        }
        String[] parameters = parametersArrayList.toArray(new String[parametersArrayList.size()]);

        if(parameterTypeLength != parameters.length) throw new DatatypeConfigurationException("Invalid number of parameters.");
        return convertParameterTypes(parameterTypes, parameters);
    }

    private Object[] convertParameterTypes(String[] parameterTypes, String[] parameters) throws DatatypeConfigurationException{
        int length = parameterTypes.length;
        Object[] parameterValues = new Object[length];
        for (int i = 0; i < length; i++) {
            parameterValues[i] = ResourceUtils.typeConversion(parameterTypes[i], parameters[i]);
        }
        return parameterValues;
    }

    private void invokeMethod(MethodInvocation methodInvocation) throws InvocationTargetException{
        try {
            Class[] parameterTypes = ResourceUtils.convertParameterTypesToClasses(methodInvocation.method.getParameters());

            Method method = this.getClass().getMethod(methodInvocation.method.getMethodName(), parameterTypes);

            Object returnValue = null;
            try {
                returnValue = method.invoke(this, methodInvocation.parameters);
                methodInvocation.executionCode = 0;
            } catch (Exception ex) {
                ex.printStackTrace();
                methodInvocation.executionCode = 1;
            }
            methodInvocation.result = returnValue;
        } catch (Exception ex) {
            throw new InvocationTargetException(ex, "Unable to execute method");
        }

    }

    private String createXMLResult(MethodInvocation[] methodInvocations) throws InvocationTargetException{
        if(methodInvocations == null || methodInvocations.length == 0) return "<invokedMethodsResult/>";
        String xml = "<invokedMethodsResult>";
        for (int i = 0; i < methodInvocations.length; i++) {
            xml += "\n" + createXMLResult(methodInvocations[i]);
        }
        xml += "\n</invokedMethodsResult>";
        return xml;
    }

    private String createXMLResult(MethodInvocation methodInvocation) throws InvocationTargetException{
        String xml = "<method name=\"" + methodInvocation.method.getMethodName() + "\" ";
        xml += "code=\"" + methodInvocation.executionCode + "\" ";
        xml += "type=\"" + methodInvocation.method.getReturnType() + "\" ";
        try {
            xml += "value=\"" + ResourceUtils.typeConversion(methodInvocation.method.getReturnType(), methodInvocation.result) + "\"/>";
        } catch (DatatypeConfigurationException ex) {
            throw new InvocationTargetException(ex, "Unable to convert result of method invocation.");
        }
        return xml;
    }

    private class MethodInvocation{
        private ResourceMethod method;
        private Object[] parameters;
        private Object result;
        private int executionCode;
    }

}
