package edu.ntnu.cikmr.services.resources;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import javax.xml.datatype.*;
import javax.xml.parsers.*;

import edu.ntnu.cikmr.*;
import net.jxta.endpoint.*;
import net.jxta.peergroup.*;
import net.jxta.pipe.*;
import net.jxta.protocol.*;
import net.jxta.util.*;
import org.w3c.dom.*;

public class Resource {
    private String peerID;
    private String peerName;
    private String pipeID;
    protected int ticker;

    private ResourceDescription resourceDescription;

    protected void setResourceDescription(ResourceDescription resourceDescription){
        if(this.resourceDescription != null){
            ResourceMethod[] methods = this.resourceDescription.getMethods();
            if(methods != null && methods.length > 0) for (int i = 0; i < methods.length; i++) methods[i].resource = null;
        }
        this.resourceDescription = resourceDescription;
        if(this.resourceDescription != null){
            ResourceMethod[] methods = this.resourceDescription.getMethods();
            if(methods != null && methods.length > 0) for (int i = 0; i < methods.length; i++) methods[i].resource = this;
        }
    }

    protected void setPeerID(String peerID) {
        this.peerID = peerID;
    }

    protected void setPeerName(String peerName) {
        this.peerName = peerName;
    }

    protected void setPipeID(String pipeID) {
        this.pipeID = pipeID;
    }

    public ResourceDescription getResourceDescription(){
        return resourceDescription;
    }

    public String getPeerID() {
        return peerID;
    }

    public String getPeerName() {
        return peerName;
    }

    public String getPipeID() {
        return pipeID;
    }

    public InvocationResult invokeMethod(ResourceMethod method, Object[] parameters) throws InvocationTargetException{
        return invokeMethods(new ResourceMethod[]{method}, new Object[][]{parameters})[0];
    }

    public InvocationResult[] invokeMethods(ResourceMethod[] methods, Object[][] parameters) throws InvocationTargetException{
        JxtaBiDiPipe pipe = null;
        InvocationResult[] response = null;
        try {
            // Create and connect to resource pipe
            PipeAdvertisement pipeAdvertisement = PipeUtilities.createPipeAdvertisement(pipeID, PipeService.UnicastType);
            pipe = new JxtaBiDiPipe();
            pipe.setReliable(true);
            pipe.connect(Framework.network.getPeerGroup(), pipeAdvertisement, 5000);
            // Send the invoke method request (Should probably have used RMI)
            pipe.sendMessage(createInvokeMethodsMessage(methods, parameters));
            // Recieve response
            Message responseMessage = pipe.getMessage(20000);
            // Extract response from message
            response = extractResults(responseMessage);
            // Close pipe
            pipe.close();
        } catch (Exception ex) {
            ex.printStackTrace();
            throw new InvocationTargetException(ex);
        }finally{
            try{
                pipe.close();
            }catch(Exception ex){}
        }
        return response;
    }

    private Message createInvokeMethodsMessage(ResourceMethod methods[], Object[][] parameters) throws DatatypeConfigurationException{
        PeerGroup peerGroup = Framework.network.getPeerGroup();

        StringMessageElement nameMessageElement = new StringMessageElement("peerName", peerGroup.getPeerName(), null);
        StringMessageElement idMessageElement = new StringMessageElement("peerID", peerGroup.getPeerID().getUniqueValue().toString(), null);
        StringMessageElement commandMessageElement = new StringMessageElement("command", "Resource:InvokeMethods", null);
        StringMessageElement xmlMessageElement = new StringMessageElement("xml", createInvokeMethodsXML(methods, parameters), null);

        Message message = new Message();
        message.addMessageElement(nameMessageElement);
        message.addMessageElement(idMessageElement);
        message.addMessageElement(commandMessageElement);
        message.addMessageElement(xmlMessageElement);

        return message;
    }

    private String createInvokeMethodsXML(ResourceMethod[] methods, Object[][] parameters) throws DatatypeConfigurationException{
        String xml = "<invokeMethods>";
        for (int i = 0; i < methods.length; i++) xml += createInvokeMethodXML(methods[i], parameters[i]);
        xml += "</invokeMethods>";
        return xml;
    }

    private String createInvokeMethodXML(ResourceMethod method, Object[] parameters) throws DatatypeConfigurationException{
        String[] parameterTypes = method.getParameters();
        int parameterTypesLength = parameterTypes == null ? 0 : parameterTypes.length;
        int parametersLength = parameters == null ? 0 : parameters.length;
        if(parameterTypesLength != parametersLength) throw new DatatypeConfigurationException("Invalid number of parameters. Expected " + parameterTypesLength + "; Found " + parametersLength + ".");

        String xml = "<method name=\"" + method.getMethodName() + "\"";
        if(parametersLength == 0){
            xml += "/>";
        }else{
            xml += ">";
            for (int i = 0; i < parametersLength; i++) {
                xml += "\n<parameter value=\"" + ResourceUtils.typeConversion(parameterTypes[i], parameters[i]) + "\"/>";
            }
            xml += "\n</method>";
        }
        return xml;
    }

    private InvocationResult[] extractResults(Message result) throws Exception{
        ByteArrayMessageElement byteArrayElement = (ByteArrayMessageElement)result.getMessageElement("command");
        String command = new String(byteArrayElement.getBytes());

        if(command.equalsIgnoreCase("Resource:InvokedMethodsResult")){
            byteArrayElement = (ByteArrayMessageElement)result.getMessageElement("xml");
            String xml = new String(byteArrayElement.getBytes());
            return extractResults(xml);
        }
        throw new Exception("Invalid invoked methods response message.");
    }

    private InvocationResult[] extractResults(String xml) throws DatatypeConfigurationException{
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(new ByteArrayInputStream(xml.getBytes()));

            NodeList nodeList = doc.getChildNodes();
            for (int i = 0; i < nodeList.getLength(); i++) {
                Node node = nodeList.item(i);
                if(node.getNodeName().equalsIgnoreCase("invokedMethodsResult")){
                    return extractResults(node);
                }
            }
        }catch(DatatypeConfigurationException ex){
            throw ex;
        }catch(Exception ex){
            throw new DatatypeConfigurationException(ex);
        }
        throw new DatatypeConfigurationException("Unable to extract method invocation results.");
    }

    private InvocationResult[] extractResults(Node result) throws DatatypeConfigurationException{
        ArrayList<InvocationResult> results = new ArrayList<InvocationResult>();

        NodeList nodeList = result.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node node = nodeList.item(i);
            if(node.getNodeName().equalsIgnoreCase("method")){
                results.add(extractResult(node));
            }
        }
        return results.toArray(new InvocationResult[results.size()]);
    }

    private InvocationResult extractResult(Node result) throws DatatypeConfigurationException{
        try {
            NamedNodeMap attributes = result.getAttributes();

            Node nameNode = attributes.getNamedItem("name");
            String name = nameNode.getNodeValue();

            Node codeNode = attributes.getNamedItem("code");
            String codeString = codeNode.getNodeValue();
            int code = Integer.parseInt(codeString, 10);

            String returnType = null;
            Object returnValue = null;
            if(code == 0){
                Node returnTypeNode = attributes.getNamedItem("type");
                returnType = returnTypeNode.getNodeValue();

                Node returnValueNode = attributes.getNamedItem("value");
                String returnValueString = returnValueNode.getNodeValue();
                returnValue = ResourceUtils.typeConversion(returnType, returnValueString);
            }
            return new InvocationResult(name, code, returnType, returnValue);
        } catch (Exception ex) {
            ex.printStackTrace();
            throw new DatatypeConfigurationException("Unable to extract method invocation result.", ex);
        }
    }

}
