package edu.ntnu.cikmr.services.resources;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.*;

import javax.xml.parsers.DocumentBuilderFactory;

import edu.ntnu.cikmr.Framework;
import edu.ntnu.cikmr.log.Log;
import edu.ntnu.cikmr.services.AbstractService;
import edu.ntnu.cikmr.services.ServiceManager;
import net.jxta.discovery.*;
import net.jxta.endpoint.*;
import net.jxta.peergroup.PeerGroup;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.util.JxtaBiDiPipe;
import org.w3c.dom.Document;
import javax.swing.event.EventListenerList;

public class ResourceManagerService extends AbstractService{
    private static final int ITERATIONS_TO_LIVE = 2;
    private static final int ITERATION_INTERVAL = 10000;
    private Hashtable<String, Resource> resources;
    private boolean isStarted;
    private ResourceServiceDiscoverer resourceServiceDiscoverer;
    private ResourceTimeoutRunner resourceTimeoutRunner;
    private PeerGroup peerGroup;
    private EventListenerList listenerList;

    public ResourceManagerService() {
        resources = new Hashtable<String,Resource>();
        listenerList = new EventListenerList();
    }

    public void addResourceListener(ResourceListener l){
        listenerList.add(ResourceListener.class, l);
    }

    public void removeResourceListener(ResourceListener l){
        listenerList.remove(ResourceListener.class, l);
    }

    private void fireResourceFound(Resource resource){
       ResourceListener[] listeners = (ResourceListener[])listenerList.getListeners(ResourceListener.class);
       ResourceEvent event = null;
       for(int i = listeners.length - 1; i >= 0; i--){
           if(event == null) event = new ResourceEvent(this, resource);
           listeners[i].resourceFound(event);
       }
    }

    private void fireResourceLost(Resource resource){
        ResourceListener[] listeners = (ResourceListener[])listenerList.getListeners(ResourceListener.class);
        ResourceEvent event = null;
        for(int i = listeners.length - 1; i >= 0; i--){
            if(event == null) event = new ResourceEvent(this, resource);
            listeners[i].resourceLost(event);
        }
    }

    /**
     * init
     *
     * @param manager ServiceManager
     * @todo Implement this edu.ntnu.cikmr.services.Service method
     */
    public void init(ServiceManager manager) {
    }

    /**
     * isStarted
     *
     * @return boolean
     * @todo Implement this edu.ntnu.cikmr.services.Service method
     */
    public boolean isStarted() {
        return isStarted;
    }

    /**
     * start
     *
     * @param peerGroup PeerGroup
     * @throws IOException
     * @todo Implement this edu.ntnu.cikmr.services.Service method
     */
    public synchronized void start(PeerGroup peerGroup) throws IOException {
        this.peerGroup = peerGroup;
        try{
            peerGroup.getDiscoveryService().flushAdvertisements("Resource:*", DiscoveryService.ADV);
        } catch(Exception ex){}
        isStarted = true;
        resourceServiceDiscoverer = new ResourceServiceDiscoverer();
        resourceTimeoutRunner = new ResourceTimeoutRunner();
        Framework.log.logMessage(this, Log.LEVEL_DEBUG, "Started.");
        fireServiceStarted(this);
    }

    /**
     * stop
     *
     * @throws IOException
     * @todo Implement this edu.ntnu.cikmr.services.Service method
     */
    public synchronized void stop() throws IOException {
        isStarted = false;
        try {
            resourceServiceDiscoverer.run = false;
            resourceServiceDiscoverer.interrupt();
            resourceServiceDiscoverer.join(5000);
        } catch (Exception ex) {}
        resourceServiceDiscoverer = null;
        try {
            resourceTimeoutRunner.run = false;
            resourceTimeoutRunner.interrupt();
            resourceTimeoutRunner.join(5000);
        } catch (Exception ex) {}
        resourceTimeoutRunner = null;
        resources.clear();
        Framework.log.logMessage(this, Log.LEVEL_DEBUG, "Stopped.");
        fireServiceStopped(this);
    }

    public Resource findResource(String type, boolean local){
        return findResource(type, null, local);
    }

    public Resource findResource(String type, String location, boolean local){
        synchronized(resources){
            Enumeration resourceEnumeration = resources.elements();
            while(resourceEnumeration.hasMoreElements()){
                Resource resource = (Resource)resourceEnumeration.nextElement();
                boolean typeMatch = (type == null);
                boolean locationMatch = (location == null);
                boolean localMatch = true;
                if(!typeMatch){
                    String resourceType = resource.getResourceDescription().getType() == null ? "" : resource.getResourceDescription().getType().getLocalName();
                    typeMatch = resourceType.equalsIgnoreCase(type);
                }
                if(!locationMatch){
                    String resourceLocation = resource.getResourceDescription().getLocation() == null ? "" : resource.getResourceDescription().getLocation().getLocalName();
                    locationMatch = resourceLocation.equalsIgnoreCase(location);
                }
                if(local){
                    String myPeerID = peerGroup.getPeerID().getUniqueValue().toString();
                    localMatch = myPeerID.equals(resource.getPeerID());
                }
                if(typeMatch && locationMatch && localMatch) return resource;
            }
        }
        return null;
    }

    public Resource[] getAll(){
        return resources.values().toArray(new Resource[]{});
    }

    private void addNewResource(PipeAdvertisement adv){
        String pipeID = adv.getPipeID().toString();
        synchronized(resources){
            Resource resource = resources.get(pipeID);
            if(resource == null){
                resource = getResource(adv);
                if(resource == null) return;
                resource.setPipeID(pipeID);
                resource.ticker = 0;
                resources.put(pipeID, resource);
                Framework.log.logMessage(this, Log.LEVEL_INFORMATIVE, "Found resource. (" + resource.getResourceDescription().getDescription() + ")");
                fireResourceFound(resource);
            }else{
                resource.ticker = 0;
            }
        }
    }

    private Resource getResource(PipeAdvertisement pipeAdvertisement){
        try {
            JxtaBiDiPipe pipe = new JxtaBiDiPipe();
            pipe.setReliable(true);
            pipe.connect(peerGroup, pipeAdvertisement, 10000);
            pipe.sendMessage(createResourceDescriptionRequestMessage(pipe.getInputPipe().getPipeID().toString()));
            Message response = pipe.getMessage(10000);
            return extractResource(response);
        } catch (IOException ex) {
            ex.printStackTrace();
            return null;
        } catch (InterruptedException ex) {
            ex.printStackTrace();
            return null;
        } catch (Exception ex) {
            ex.printStackTrace();
            return null;
        }
    }

    private Resource extractResource(Message message){
        try {
            ByteArrayMessageElement byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("peerName");
            String name = new String(byteArrayElement.getBytes());
            byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("peerID");
            String peerID = new String(byteArrayElement.getBytes());
            byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("command");
            String command = new String(byteArrayElement.getBytes());

            if(!command.equalsIgnoreCase("Resource:Description")) return null;

            byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("xml");
            String xml = new String(byteArrayElement.getBytes());

            Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(new ByteArrayInputStream(xml.getBytes()));
            ResourceDescription resourceDescription = ResourceDescription.fromXML(document);

            if(resourceDescription == null) return null;

            Resource resource = new Resource();
            resource.setPeerID(peerID);
            resource.setPeerName(name);
            resource.setResourceDescription(resourceDescription);

            return resource;
        } catch (Exception ex) {
            ex.printStackTrace();
            Framework.log.logMessage(this, Log.LEVEL_ERROR, "Unable to decode recieved message.");
        }
        return null;
    }

    private Message createResourceDescriptionRequestMessage(String pipeID){
        StringMessageElement nameMessageElement = new StringMessageElement("peerName", peerGroup.getPeerName(), null);
        StringMessageElement idMessageElement = new StringMessageElement("peerID", peerGroup.getPeerID().getUniqueValue().toString(), null);
        StringMessageElement pipeIDMessageElement = new StringMessageElement("pipeID", pipeID, null);
        StringMessageElement commandMessageElement = new StringMessageElement("command", "Resource:GetDescription", null);

        Message message = new Message();
        message.addMessageElement(nameMessageElement);
        message.addMessageElement(idMessageElement);
        message.addMessageElement(pipeIDMessageElement);
        message.addMessageElement(commandMessageElement);
        return message;
    }

    private void resourceRemoved(Resource resource){
        Framework.log.logMessage(ResourceManagerService.this, Log.LEVEL_INFORMATIVE, "Lost resource. (" + resource.getResourceDescription().getDescription() + ")");
        fireResourceLost(resource);
    }

    private class ResourceServiceDiscoverer extends Thread implements DiscoveryListener{
        private boolean run;

        public ResourceServiceDiscoverer(){
            run = true;
            start();
        }

        public void run(){
            Framework.log.logMessage(ResourceManagerService.this, Log.LEVEL_DEBUG, "Discovery runner started.");
            while(run){
                try {
                    discoverLocalResources();
                    peerGroup.getDiscoveryService().getRemoteAdvertisements(null, DiscoveryService.ADV, PipeAdvertisement.NameTag, "Resource:*", 30, this);
                    Framework.log.logMessage(ResourceManagerService.this, Log.LEVEL_INFORMATIVE, "Sending discovery request.");
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
                try {
                    Thread.sleep(ITERATION_INTERVAL);
                } catch (Exception ex) {}
            }
            Framework.log.logMessage(ResourceManagerService.this, Log.LEVEL_DEBUG, "Discovery runner stopped.");
        }

        public void discoverLocalResources(){
            try {
                Enumeration e = peerGroup.getDiscoveryService().getLocalAdvertisements(DiscoveryService.ADV, PipeAdvertisement.NameTag, "Resource:*");
                while (e.hasMoreElements()) {
                    try {
                        PipeAdvertisement pipeAdv = (PipeAdvertisement) e.nextElement();
                        if(!pipeAdv.getName().startsWith("Resource:")) continue;
                        addNewResource(pipeAdv);
                    } catch (Exception ex) {}
                }
            } catch (Exception ex) {}
        }

        public void discoveryEvent(DiscoveryEvent discoveryEvent) {
            Enumeration e = discoveryEvent.getSearchResults();
            while(e.hasMoreElements()){
                try {
                    PipeAdvertisement pipeAdv = (PipeAdvertisement) e.nextElement();
                    peerGroup.getDiscoveryService().flushAdvertisement(pipeAdv);
                    if(!pipeAdv.getName().startsWith("Resource:")) continue;
                    addNewResource(pipeAdv);
                } catch (Exception ex) {}
            }
        }

    }

    private class ResourceTimeoutRunner extends Thread{
        private boolean run;

        public ResourceTimeoutRunner(){
            run = true;
            start();
        }

        public void run(){
            Framework.log.logMessage(ResourceManagerService.this, Log.LEVEL_DEBUG, "Timeout runner started.");
            while(run){
                try {
                    synchronized(resources){
                        Iterator<Resource> iterator = resources.values().iterator();
                        while(iterator.hasNext()){
                            Resource resource = iterator.next();
                            if(resource.ticker == ITERATIONS_TO_LIVE){
                                iterator.remove();
                                resourceRemoved(resource);
                            } else{
                                resource.ticker++;
                            }
                        }
                    }
                } catch (Exception ex) {
                    ex.printStackTrace();
                }

                try{
                    Thread.sleep(ITERATION_INTERVAL);
                } catch(Exception ex){}
            }
            Framework.log.logMessage(ResourceManagerService.this, Log.LEVEL_DEBUG, "Timeout runner stopped.");
        }

    }

}
