package edu.ntnu.cikmr.services.resources;

import java.io.*;
import java.text.*;
import java.util.*;

import javax.xml.parsers.*;

import org.w3c.dom.*;
import javax.xml.datatype.*;

public class ResourceUtils {
    private static int uniqueID;
    private static Hashtable<String,DataType> parameterDataTypes;

    static{
        uniqueID = Integer.MIN_VALUE;
        parameterDataTypes = new Hashtable<String,DataType>(3);
        registerDataType(new IntegerDataType());
        registerDataType(new DoubleDataType());
        registerDataType(new StringDataType());
        registerDataType(new VoidDataType());
    }

    public static String generateUniqueID(){
        return Integer.toHexString(uniqueID++).toUpperCase();
    }

    public static void registerDataType(DataType dataType){
        parameterDataTypes.put(dataType.getDataTypeName(), dataType);
    }

    public static boolean hasDataType(String name){
        return parameterDataTypes.containsKey(name);
    }

    public static DataType removeDataType(String name){
        return parameterDataTypes.remove(name);
    }

    public static DataType getDataType(String name){
        return parameterDataTypes.get(name);
    }

    public static Class[] convertParameterTypesToClasses(String[] types) throws DatatypeConfigurationException{
        if(types == null || types.length == 0) return null;
        Class[] parameters = new Class[types.length];
        for(int i = 0; i < types.length; i++){
            String type = types[i];
            Class typeClass = parameterDataTypes.get(type).getDataTypeClass();
            if(typeClass == null) throw new DatatypeConfigurationException("The type \"" + type + "\" is not a known parameter type.");
            parameters[i] = typeClass;
        }
        return parameters;
    }

    public static ResourceDescription getResourceDescriptionFromXML(String xml){
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(new ByteArrayInputStream(xml.getBytes()));
            return getResourceDescriptionFromXML(doc);
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return null;
    }


    public static ResourceDescription getResourceDescriptionFromXML(File file){
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(file);
            return getResourceDescriptionFromXML(doc);
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return null;
    }

    private static ResourceDescription getResourceDescriptionFromXML(Document doc){
        try {
            NodeList nodeList = doc.getChildNodes();
            for (int i = 0; i < nodeList.getLength(); i++) {
                Node node = nodeList.item(i);
                if(node.getNodeName().equalsIgnoreCase("resource")){
                    try {
                        return ResourceDescription.fromXML(node);
                    } catch (ParseException ex) {
                        ex.printStackTrace();
                        return null;
                    }
                }
            }
        } catch (Exception ex) {
            ex.printStackTrace();
            return null;
        }
        return null;
    }

    public static String typeConversion(String type, Object value) throws DatatypeConfigurationException{
        DataType dataType = getDataType(type);
        if(dataType == null) throw new DatatypeConfigurationException("DataType is not registered.");
        return dataType.convertDataType(value);
    }

    public static Object typeConversion(String type, String value) throws DatatypeConfigurationException{
        DataType dataType = getDataType(type);
        if(dataType == null) throw new DatatypeConfigurationException("DataType is not registered.");
        return dataType.convertDataType(value);
    }

    public static void printResourceDescription(ResourceDescription resourceDescription, OutputStream outputStream){
        PrintStream printStream = new PrintStream(outputStream);

        printStream.println("--- Resource description ---");
        printStream.println("\tType: " + resourceDescription.getType().getLocalName());
        printStream.println("\tDescription: " + resourceDescription.getDescription());
        printStream.println("\tLocation: " + resourceDescription.getLocation().getLocalName());
        ResourceMethod[] rsm = resourceDescription.getMethods();
        for (int i = 0; i < rsm.length; i++) {
            printStream.print("\tMethod: " + rsm[i].getMethodName() + "(");
            String[] params = rsm[i].getParameters();
            for (int p = 0; p < params.length; p++) {
                printStream.print((p == 0 ? "" : ", ") + params[i]);
            }
            printStream.println(") : " + rsm[i].getReturnType());
        }
        printStream.println("-----------------------------");
    }

    private static class IntegerDataType implements DataType{
        public Class getDataTypeClass() {
            return int.class;
        }

        public Object convertDataType(String value) {
            try{
                return new Integer(value);
            }catch(Exception ex){}
            return new Integer(0);
        }

        public String convertDataType(Object value) {
            if(value == null) return "0";
            return value.toString();
        }

        public String getDataTypeName() {
            return "int";
        }
    }

    private static class DoubleDataType implements DataType{
        public Class getDataTypeClass() {
            return double.class;
        }

        public Object convertDataType(String value) {
            try{
                return new Double(value);
            }catch(Exception ex){}
            return new Double(0);
        }

        public String convertDataType(Object value) {
            if(value == null) return "0.0";
            return value.toString();
        }

        public String getDataTypeName() {
            return "double";
        }
    }

    private static class StringDataType implements DataType{
        public Class getDataTypeClass() {
            return String.class;
        }

        public Object convertDataType(String value) {
            boolean isNullValue = value == null;
            if(!isNullValue) isNullValue = value.equals("null");
            return isNullValue ? null : value;
        }

        public String convertDataType(Object value) {
            return value == null ? "null" : value.toString();
        }

        public String getDataTypeName() {
            return "String";
        }
    }

    private static class VoidDataType implements DataType{
        public Class getDataTypeClass() {
            return void.class;
        }

        public Object convertDataType(String value) {
            return null;
        }

        public String convertDataType(Object value) {
            return "";
        }

        public String getDataTypeName() {
            return "void";
        }
    }

}
