package edu.ntnu.cikmr.workflowenactment;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import edu.ntnu.cikmr.Framework;
import net.jxta.endpoint.*;
import net.jxta.peergroup.PeerGroup;
import net.jxta.pipe.PipeService;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.util.JxtaBiDiPipe;
import net.jxta.util.PipeUtilities;
import org.w3c.dom.*;

public class RemoteWorkflowConnector{
    private String pipeID;
    private Workflow workflow;
    private UpdateThread updateThread;
    private String logCode;
    private int failures;

    public RemoteWorkflowConnector(String pipeID, Workflow workflow){
        this.pipeID = pipeID;
        this.workflow = workflow;
        logCode = "0";
    }

    public void start(){
        if(updateThread != null && updateThread.run) return;
        failures = 0;
        updateThread = new UpdateThread();
    }

    public void stop(){
        if(updateThread != null && updateThread.run){
            updateThread.run = false;
            try{
                updateThread.interrupt();
                updateThread.join();
            } catch(Exception ex){}
            updateThread = null;
        }
    }

    public void leaveWorkflow(){
        stop();

        if(!Framework.network.getConnectionStatus().isConnected()) return;

        JxtaBiDiPipe pipe = null;
        try {
            // Create and connect to resource pipe
            PipeAdvertisement pipeAdvertisement = PipeUtilities.createPipeAdvertisement(pipeID, PipeService.UnicastType);
            pipe = new JxtaBiDiPipe();
            pipe.setReliable(true);
            pipe.connect(Framework.network.getPeerGroup(), pipeAdvertisement, 5000);
            pipe.sendMessage(createLeaveWorkflowMessage());
        } catch (Exception ex) {
            ex.printStackTrace();
        }finally{
            try{
                pipe.close();
            }catch(Exception ex){
                ex.printStackTrace();
            }
        }
    }

    private Message createLeaveWorkflowMessage(){
        PeerGroup peerGroup = Framework.network.getPeerGroup();

        StringMessageElement nameMessageElement = new StringMessageElement("peerName", peerGroup.getPeerName(), null);
        StringMessageElement idMessageElement = new StringMessageElement("peerID", peerGroup.getPeerID().getUniqueValue().toString(), null);
        StringMessageElement commandMessageElement = new StringMessageElement("command", "Workflow:Leave", null);

        Message message = new Message();
        message.addMessageElement(nameMessageElement);
        message.addMessageElement(idMessageElement);
        message.addMessageElement(commandMessageElement);

        return message;
    }

    private void update(){
        if(failures++ == 2){
            new Thread(){
                public void run(){
                    workflow.lostContact();
                }
            }.start();
            return;
        }

        if(!Framework.network.getConnectionStatus().isConnected()) return;

        JxtaBiDiPipe pipe = null;
        try {
            // Create and connect to resource pipe
            PipeAdvertisement pipeAdvertisement = PipeUtilities.createPipeAdvertisement(pipeID, PipeService.UnicastType);
            pipe = new JxtaBiDiPipe();
            pipe.setReliable(true);
            pipe.connect(Framework.network.getPeerGroup(), pipeAdvertisement, 5000);
            pipe.sendMessage(createUpdateRequestMessage());
            Message responseMessage = pipe.getMessage(20000);
            if(responseMessage != null){
                handleUpdateResponseMessage(responseMessage);
                failures = 0;
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }finally{
            try{
                pipe.close();
            }catch(Exception ex){
                ex.printStackTrace();
            }
        }
    }

    private Message createUpdateRequestMessage(){
        PeerGroup peerGroup = Framework.network.getPeerGroup();

        StringMessageElement nameMessageElement = new StringMessageElement("peerName", peerGroup.getPeerName(), null);
        StringMessageElement idMessageElement = new StringMessageElement("peerID", peerGroup.getPeerID().getUniqueValue().toString(), null);
        StringMessageElement commandMessageElement = new StringMessageElement("command", "Workflow:Status", null);
        StringMessageElement logCodeMessageElement = new StringMessageElement("logCode", logCode, null);

        Message message = new Message();
        message.addMessageElement(nameMessageElement);
        message.addMessageElement(idMessageElement);
        message.addMessageElement(commandMessageElement);
        message.addMessageElement(logCodeMessageElement);

        return message;
    }

    private void handleUpdateResponseMessage(Message message){
        try{
            ByteArrayMessageElement byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("command");
            String command = new String(byteArrayElement.getBytes());

            if(command.equalsIgnoreCase("Workflow:Status")){
                byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("xml");
                String xml = new String(byteArrayElement.getBytes());
                handleUpdateResponseXML(xml);
            }
        } catch(Exception ex){
            ex.printStackTrace();
        }
    }

    private void handleUpdateResponseXML(String xml){
        try{
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(new ByteArrayInputStream(xml.getBytes()));

            NodeList nodeList = doc.getChildNodes();
            for (int i = 0; i < nodeList.getLength(); i++) {
                Node node = nodeList.item(i);
                if(node.getNodeName().equalsIgnoreCase("workflowStatus")){
                    handleUpdateResponseXML(node);
                }
            }
        }catch(Exception ex){
            ex.printStackTrace();
        }
    }

    private void handleUpdateResponseXML(Node n){
        NamedNodeMap map = n.getAttributes();
        Node running = map.getNamedItem("running");
        boolean isRunning = running == null ? false : new Boolean(running.getNodeValue()).booleanValue();

        workflow.setIsExecuting(isRunning);

        NodeList nodeList = n.getChildNodes();
        for(int i = 0; i < nodeList.getLength(); i++){
            Node childNode = nodeList.item(i);
            if(childNode.getNodeName().equalsIgnoreCase("participants")){
                updateParticipants(childNode);
            }else if(childNode.getNodeName().equalsIgnoreCase("log")){
                updateLog(childNode);
            }
        }
    }

    private void updateParticipants(Node n){
        ArrayList<Participant> participants = new ArrayList<Participant>();
        NodeList nodeList = n.getChildNodes();
        for(int i = 0; i < nodeList.getLength(); i++){
            Node childNode = nodeList.item(i);
            if(childNode.getNodeName().equalsIgnoreCase("participant")) extractParticipant(childNode, participants);
        }
        if(participants.size() > 0){
            workflow.synchronizeParticipants(participants);
        }
    }

    private void extractParticipant(Node n, ArrayList<Participant> participants){
        NamedNodeMap map = n.getAttributes();
        Node peerName = map.getNamedItem("peerName");
        String speerName = peerName == null ? null : peerName.getNodeValue();
        Node peerID = map.getNamedItem("peerID");
        String speerID = peerID == null ? null : peerID.getNodeValue();

        if(speerName == null || speerID == null) return;
        participants.add(new Participant(speerID, speerName));
    }

    private void updateLog(Node n){
        ArrayList<WorkflowLogEntry> entries = new ArrayList<WorkflowLogEntry>();

        NamedNodeMap map = n.getAttributes();
        Node code = map.getNamedItem("code");
        logCode = code == null ? logCode : code.getNodeValue();

        NodeList nodeList = n.getChildNodes();
        for(int i = 0; i < nodeList.getLength(); i++){
            Node childNode = nodeList.item(i);
            if(childNode.getNodeName().equalsIgnoreCase("entry")){
                WorkflowLogEntry entry = getLogEntry(childNode);
                if(entry != null) entries.add(entry);
            }
        }
        if(entries.size() > 0) workflow.getLog().addEntries(entries.toArray(new WorkflowLogEntry[entries.size()]));
    }

    private WorkflowLogEntry getLogEntry(Node n){
        NamedNodeMap map = n.getAttributes();
        Node nodeActivity = map.getNamedItem("activity");
        String activity = nodeActivity == null ? "" : nodeActivity.getNodeValue();
        Node nodeStatement = map.getNamedItem("statement");
        String statement = nodeStatement == null ? "" : nodeStatement.getNodeValue();

        Node nodeLevel = map.getNamedItem("level");
        int level = 0;
        try{
            level = Integer.parseInt(nodeLevel.getNodeValue());
        } catch(Exception ex){}

        String message = getTextFromNode(n);

        return new WorkflowLogEntry(activity, statement, level, message);
    }

    private static String getTextFromNode(Node node){
        String text = "";
        NodeList nodeList = node.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            Node childNode = nodeList.item(i);
            if(childNode.getNodeType() == Node.TEXT_NODE) text += childNode.getNodeValue();
        }
        return text;
    }

    private class UpdateThread extends Thread{
        private boolean run;

        public UpdateThread(){
            run = true;
            start();
        }

        public void run(){
            while(run){
                update();
                try{
                    Thread.sleep(5000);
                } catch(Exception ex){}
            }
        }

    }

}
