package edu.ntnu.cikmr.workflowenactment;

import java.net.SocketTimeoutException;
import java.util.EventObject;

import edu.ntnu.cikmr.Framework;
import edu.ntnu.cikmr.services.resources.ResourceUtils;
import net.jxta.endpoint.Message;
import net.jxta.peergroup.PeerGroup;
import net.jxta.pipe.PipeService;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.util.*;
import net.jxta.endpoint.StringMessageElement;
import net.jxta.endpoint.ByteArrayMessageElement;
import java.util.Iterator;

public class RemoteWorkflowServer implements WorkflowLogListener{
    private Workflow workflow;
    private int logCode;
    private JxtaServerPipe serverPipe;
    private ServerPipeRunner serverPipeRunner;

    public RemoteWorkflowServer(Workflow workflow){
        this.workflow = workflow;
        workflow.getLog().addWorkflowLogListener(this);

    }

    public boolean start(){
        if(!Framework.network.getConnectionStatus().isConnected()) return false;
        if(serverPipeRunner != null && serverPipeRunner.run) return true;
        try{
            PeerGroup peerGroup = Framework.network.getPeerGroup();
            PipeAdvertisement pipeAdvertisement = PipeUtilities.createNewPipeAdvertisement(peerGroup, PipeService.UnicastType);
            pipeAdvertisement.setName("Workflow:" + peerGroup.getPeerID().getUniqueValue().toString() + ":" + workflow.getID());
            peerGroup.getDiscoveryService().publish(pipeAdvertisement);
            serverPipe = new JxtaServerPipe(peerGroup, pipeAdvertisement, 20, 2000);
            serverPipeRunner = new ServerPipeRunner();
        }catch(Exception ex){
            ex.printStackTrace();
            return false;
        }
        return true;
    }

    public void stop(){
        if(serverPipeRunner != null && serverPipeRunner.run){
            serverPipeRunner.run = false;
            try{
                serverPipeRunner.interrupt();
                serverPipeRunner.join();
            } catch(Exception ex){}
            try{
                serverPipe.close();
            } catch(Exception ex){
                ex.printStackTrace();
            }
            serverPipe = null;
            serverPipeRunner = null;
        }
    }

    public String getPipeID(){
        return serverPipe == null ? null : serverPipe.getPipeAdv().getPipeID().toString();
    }

    public void logUpdated(EventObject e){
        logCode++;
    }

    private class ServerPipeRunner extends Thread{
        private boolean run;
        private JxtaBiDiPipe pipe;

        public ServerPipeRunner(){
            run = true;
            start();
        }

        public void run(){
            while(run){
                try {
                    pipe = serverPipe.accept();
                } catch (SocketTimeoutException ex) {
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
                if(pipe != null){
                    new ConnectionHandler(pipe);
                    pipe = null;
                }
            }
        }

    }

    private class ConnectionHandler extends Thread{
        private JxtaBiDiPipe pipe;

        private ConnectionHandler(JxtaBiDiPipe pipe){
            this.pipe = pipe;
            start();
        }

        public void run(){
            try {
                Message requestMessage = pipe.getMessage(10000);
                Message responseMessage = createResponse(requestMessage);
                if(responseMessage != null) pipe.sendMessage(responseMessage);
                try {
                    pipe.close();
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }

    }

    private Message createResponse(Message message){
        ByteArrayMessageElement byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("peerID");
        String peerID = new String(byteArrayElement.getBytes());
        byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("command");
        String command = new String(byteArrayElement.getBytes());
        try{
            if(command.equalsIgnoreCase("Workflow:Status")){
                byteArrayElement = (ByteArrayMessageElement)message.getMessageElement("logCode");
                String logCode = new String(byteArrayElement.getBytes());
                // Only return a result if the peer is part of the workflow
                return workflow.hasParticipant(peerID) ? createWorkflowStatusMessage(logCode) : null;
            }else if(command.equalsIgnoreCase("Workflow:Leave")){
                if(workflow.hasParticipant(peerID)){
                    workflow.removeParticipant(peerID);
                }
                // For now; Do not send a response
                return null;
            }
            return null;
        } catch(Exception ex){
            ex.printStackTrace();
            return null;
        }
    }

    private Message createWorkflowStatusMessage(String logCode){
        PeerGroup peerGroup = Framework.network.getPeerGroup();

        StringMessageElement nameMessageElement = new StringMessageElement("peerName", peerGroup.getPeerName(), null);
        StringMessageElement idMessageElement = new StringMessageElement("peerID", peerGroup.getPeerID().getUniqueValue().toString(), null);
        StringMessageElement commandMessageElement = new StringMessageElement("command", "Workflow:Status", null);
        StringMessageElement xmlMessageElement = new StringMessageElement("xml", createWorkflowStatusXML(logCode), null);

        Message message = new Message();
        message.addMessageElement(nameMessageElement);
        message.addMessageElement(idMessageElement);
        message.addMessageElement(commandMessageElement);
        message.addMessageElement(xmlMessageElement);

        return message;
    }

    private String createWorkflowStatusXML(String logCode){
        String xml = "<workflowStatus running=\"" + workflow.isExecuting() + "\">\n";
        xml += createParticipantsXML() + "\n";
        xml += createLogXML(logCode) + "\n";
        xml += "</workflowStatus>";
        return xml;
    }

    private String createParticipantsXML(){
        String xml = "<participants>";
        for(int i = 0; i < workflow.getParticipantsCount(); i++){
            Participant p = workflow.getParticipant(i);
            xml += "<participant peerName=\"" + p.getPeerName() + "\" peerID=\"" + p.getPeerID() + "\"/>";
        }
        xml += "</participants>";
        return xml;
    }

    private String createLogXML(String logCode){
        int code = 0;
        try{
            code = Integer.parseInt(logCode, 10);
        } catch(Exception ex){}

        int diff = this.logCode - code;
        if(diff > 0){
            if(diff > 100) diff = 100;
            if(diff > workflow.getLog().getEntriesCount()) diff = workflow.getLog().getEntriesCount();
            String xml = "<log code=\"" + this.logCode + "\">\n";
            for(int i = 0; i < diff; i++){
                xml += createWorkflowLogEntryXML(workflow.getLog().getEntry(i));
            }
            xml += "</log>";
            return xml;
        }else{
            return "";
        }
    }

    private String createWorkflowLogEntryXML(WorkflowLogEntry e){
        return "<entry activity=\"" + (e.getActivity() == null ? "" : e.getActivity()) + "\" statement=\"" + (e.getStatement() == null ? "" : e.getStatement()) + "\" level=\"" + e.getLevel() + "\">" + (e.getMessage() == null ? "" : e.getMessage()) + "</entry>";
    }

}
