package edu.ntnu.cikmr.workflowenactment;

import java.io.ByteArrayInputStream;
import java.util.*;

import javax.swing.event.EventListenerList;
import javax.xml.parsers.DocumentBuilderFactory;

import edu.ntnu.cikmr.workflow.Goal;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import edu.ntnu.cikmr.Framework;
import edu.ntnu.cikmr.services.resources.*;
import edu.ntnu.cikmr.services.location.*;
import java.util.Map.Entry;

public class Workflow implements WorkflowControlListener{
    private EventListenerList listenerList;
    private boolean executing;
    private boolean executedLocally;
    private String id;
    private Goal goal;

    private WorkflowLog log;
    private List<Participant> participants;

    private WorkflowContext context;
    private RemoteWorkflowConnector connector;
    private RemoteWorkflowServer server;


    public Workflow(String xml) throws IllegalArgumentException{
         if(xml == null || xml.trim().length() == 0) throw new IllegalArgumentException("A valid workflow definition must be supplied.");
         executedLocally = true;

         listenerList = new EventListenerList();
         log = new SimpleWorkflowLog();
         participants = Collections.synchronizedList(new ArrayList<Participant>());
         addLocalParticipant();


         goal = new Goal();

         context = new WorkflowContext();
         context.setLog(log);
         context.setGoal(goal);
         context.addWorkflowControlListener(this);

         try{
             ByteArrayInputStream inputStream = new ByteArrayInputStream(xml.getBytes());
             Document document = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(inputStream);
             NodeList nodeList = document.getElementsByTagName("goal");
             if(nodeList.getLength() == 1){
                 goal.fromXML(nodeList.item(0));
             }else{
                 throw new Exception("Invalid workflow xml.");
             }
         } catch(Exception ex){
             ex.printStackTrace();
             throw new IllegalArgumentException("Invalid workflow xml.");
         }

         id = ResourceUtils.generateUniqueID();

         server = new RemoteWorkflowServer(this);
         server.start();
     }

    public boolean isExecuting(){
        return executing;
    }

    public String getID(){
        return id;
    }

    protected void setIsExecuting(boolean b){
        if(b != executing){
            executing = b;
            if(executing){
                fireWorkflowStarted();
            }else{
                fireWorkflowFinished();
            }
        }
    }

    public boolean isExecutedLocally(){
        return executedLocally;
    }

    public WorkflowLog getLog(){
        return log;
    }

    public void addWorkflowLifecycleListener(WorkflowLifecycleListener l){
        listenerList.add(WorkflowLifecycleListener.class, l);
    }

    public void removeWorkflowLifecycleListener(WorkflowLifecycleListener l){
        listenerList.remove(WorkflowLifecycleListener.class, l);
    }

    private void fireWorkflowStarted(){
        WorkflowEvent event = new WorkflowEvent(this, this);
        WorkflowLifecycleListener[] listeners = listenerList.getListeners(WorkflowLifecycleListener.class);
        for(int i = listeners.length - 1; i >= 0; i--) listeners[i].workflowStarted(event);
    }

    private void fireWorkflowStopped(){
        WorkflowEvent event = new WorkflowEvent(this, this);
        WorkflowLifecycleListener[] listeners = listenerList.getListeners(WorkflowLifecycleListener.class);
        for(int i = listeners.length - 1; i >= 0; i--) listeners[i].workflowStopped(event);
    }

    private void fireWorkflowFinished(){
        WorkflowEvent event = new WorkflowEvent(this, this);
        WorkflowLifecycleListener[] listeners = listenerList.getListeners(WorkflowLifecycleListener.class);
        for(int i = listeners.length - 1; i >= 0; i--) listeners[i].workflowFinished(event);
    }

    private void fireWorkflowExecutionMoved(){
        WorkflowEvent event = new WorkflowEvent(this, this);
        WorkflowLifecycleListener[] listeners = listenerList.getListeners(WorkflowLifecycleListener.class);
        for(int i = listeners.length - 1; i >= 0; i--) listeners[i].workflowExecutionMoved(event);
    }

    private void fireWorkflowParticipantsChanged(){
        WorkflowEvent event = new WorkflowEvent(this, this);
        WorkflowLifecycleListener[] listeners = listenerList.getListeners(WorkflowLifecycleListener.class);
        for(int i = listeners.length - 1; i >= 0; i--) listeners[i].workflowParticipantsChanged(event);
    }

    public void addParticipant(String peerID, String peerName){
        addParticipant(new Participant(peerID, peerName));
    }

    public void addParticipant(Participant p){
        if(!participants.contains(p)){
            participants.add(p);
            fireWorkflowParticipantsChanged();
        }
    }

    public void removeParticipant(String peerID){
        if(participants.remove(new Participant(peerID, null))) fireWorkflowParticipantsChanged();
    }

    public boolean hasParticipant(String peerID){
        return participants.contains(new Participant(peerID, null));
    }

    public int getParticipantsCount(){
        return participants.size();
    }

    public Participant getParticipant(int index){
        return participants.get(index);
    }

    public void synchronizeParticipants(List<Participant> p){
        participants.clear();
        participants.addAll(p);
        fireWorkflowParticipantsChanged();
    }

    public void execute(){
        if(executedLocally) new ExecuteThread();
    }

    public void halt(){
        if(executedLocally) context.setHalted(true);
    }

    public void setDefined(String name, String value){
        if(executedLocally){
            context.getGoal().setDefinition(name, value);
        }
    }

    public Goal getGoal(){
        return goal;
    }

    public void controlCommand(WorkflowControlEvent e){
        if(executedLocally){
            if(e.getCommand().equalsIgnoreCase("halt")){
                halt();
            }
        }
    }

    private class ExecuteThread extends Thread{
        private ExecuteThread(){
            context.setHalted(false);
            start();
        }

        public void run(){
            executing = true;
            fireWorkflowStarted();
            if(context.execute()){
                executing = false;
                fireWorkflowFinished();
            }else{
                executing = false;
                fireWorkflowStopped();
            }
        }

    }

    // Remote things

    public void remoteConnect(String pipeID){
        if(executedLocally){
            if(executing){
                halt();
                lockUntilFinished();
            }
            executedLocally = false;
            connector = new RemoteWorkflowConnector(pipeID, this);
            connector.start();
            fireWorkflowExecutionMoved();
        }
    }

    public void moveLocal(){
        if(!executedLocally){
            if(connector != null){
                connector.leaveWorkflow();
                connector.stop();
                connector = null;
                setIsExecuting(false);
            }
            executedLocally = true;
            fireWorkflowExecutionMoved();
            addLocalParticipant();
        }
    }

    private void addLocalParticipant(){
        if(Framework.network.getConnectionStatus().isConnected()){
            participants.clear();
            addParticipant(Framework.network.getPeerGroup().getPeerID().getUniqueValue().toString(), Framework.network.getPeerGroup().getPeerName());
        }
    }

    protected void lostContact(){
        moveLocal();
    }

    public void release(){
        if(executing){
            halt();
            lockUntilFinished();
        }
        server.stop();
        server = null;
        if(connector != null && !executedLocally){
            connector.leaveWorkflow();
            connector.stop();
            connector = null;
        }
    }

    private void lockUntilFinished(){
        while(executing && executedLocally){
            try{
                Thread.sleep(100);
            } catch(Exception ex){}
        }
    }

    public String getServerPipeID(){
        return server.getPipeID();
    }

    public String toTypeXML(){
        if(goal == null) return null;
        LocationSensorResource location = (LocationSensorResource)Framework.network.getServiceManager().get(LocationSensorResource.class);
        return goal.toTypeXML(location == null ? null : location.getLocation(), getServerPipeID(), id);
    }

}
